# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import cmath
import typing
from typing import List, Tuple, Union

from ..utils import check_unitarity
from ..gate import OneQubitGate, RYGate, RZGate, UGate

if typing.TYPE_CHECKING:
    import numpy as np


def calc_uparams(mat: 'np.ndarray') -> Tuple[float, float, float, float]:
    """Calculate U-gate parameters from a 2x2 unitary matrix."""
    assert mat.shape == (2, 2)
    assert check_unitarity(mat)
    gamma = cmath.phase(mat[0, 0])
    mat = mat * cmath.exp(-1j * gamma)
    theta = math.atan2(abs(mat[1, 0]), mat[0, 0].real) * 2.0
    phi_plus_lambda = cmath.phase(mat[1, 1])
    phi = cmath.phase(mat[1, 0]) % (2.0 * math.pi)
    lam = (phi_plus_lambda - phi) % (2.0 * math.pi)
    return theta, phi, lam, gamma


def u_decomposer(gate: OneQubitGate) -> List[UGate]:
    """Decompose one qubit gate to U gate.

    Args:
        gate (OneQubitGate): The gate which have 2x2 unitary matrix

    Returns:
        List of U gate. length is always 1.
    """
    mat = gate.matrix()
    theta, phi, lam, gamma = calc_uparams(mat)
    return [UGate.create(gate.targets, (theta, phi, lam, gamma))]


def ryrz_decomposer(gate: OneQubitGate) -> List[Union[RYGate, RZGate]]:
    """Decompose one qubit gate to RY and RZ gate.

    Args:
        gate (OneQubitGate): The gate which have 2x2 unitary matrix

    Returns:
        List of RZ and RY gate. The global phase is omitted.
    """
    mat = gate.matrix()
    theta, phi, lam, _ = calc_uparams(mat)
    if theta < 1e-10:
        if phi + lam < 1e-10:
            return []
        return [RZGate.create(gate.targets, (phi + lam,))]
    ops = []
    if 1e-10 < lam < 2.0 * math.pi - 1e-10:
        ops.append(RZGate.create(gate.targets, (lam,)))
    ops.append(RYGate(gate.targets, theta))
    if 1e-10 < phi < 2.0 * math.pi - 1e-10:
        ops.append(RZGate.create(gate.targets, (phi,)))
    return ops



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for convenient."""
from collections import Counter
import typing
from typing import Dict, Tuple, Union

import numpy as np

if typing.TYPE_CHECKING:
    from . import Circuit


def to_inttuple(
    bitstr: Union[str, Counter, Dict[str, int]]
) -> Union[Tuple, Counter, Dict[Tuple, int]]:
    """Convert from bit string likes '01011' to int tuple likes (0, 1, 0, 1, 1)

    Args:
        bitstr (str, Counter, dict): String which is written in "0" or "1".
            If all keys are bitstr, Counter or dict are also can be converted by this function.

    Returns:
        tuple of int, Counter, dict: Converted bits.
            If bitstr is Counter or dict, returns the Counter or dict
            which contains {converted key: original value}.

    Raises:
        ValueError: If bitstr type is unexpected or bitstr contains illegal character.
    """
    if isinstance(bitstr, str):
        return tuple(int(b) for b in bitstr)
    if isinstance(bitstr, Counter):
        return Counter(
            {tuple(int(b) for b in k): v
             for k, v in bitstr.items()})
    if isinstance(bitstr, dict):
        return {tuple(int(b) for b in k): v for k, v in bitstr.items()}
    raise ValueError("bitstr type shall be `str`, `Counter` or `dict`")


def ignore_global_phase(statevec: np.ndarray) -> np.ndarray:
    """Multiple e^-iθ to `statevec` where θ is a phase of first non-zero element.

    Args:
        statevec np.ndarray: Statevector.

    Returns:
        np.ndarray: `statevec` is returned.
    """
    for q in statevec:
        if abs(q) > 0.0000001:
            ang = abs(q) / q
            statevec *= ang
            break
    return statevec


def check_unitarity(mat: np.ndarray) -> bool:
    """Check whether mat is a unitary matrix."""
    shape = mat.shape
    if len(shape) != 2 or shape[0] != shape[1]:
        # Not a square matrix.
        return False
    return np.allclose(mat @ mat.T.conjugate(), np.eye(shape[0]))


def circuit_to_unitary(circ: 'Circuit', *runargs, **runkwargs):
    """Make circuit to unitary. This function is experimental feature and
    may changed or deleted in the future."""

    # To avoid circuilar import, import here.
    from . import Circuit

    runkwargs.setdefault('returns', 'statevector')
    runkwargs.setdefault('ignore_global', False)
    n_qubits = circ.n_qubits
    vecs = []
    if n_qubits == 0:
        return np.array([[1]])
    for i in range(1 << n_qubits):
        bitmask = tuple(k for k in range(n_qubits) if (1 << k) & i)
        c = Circuit()
        if bitmask:
            c.x[bitmask]
        c += circ
        vecs.append(c.run(*runargs, **runkwargs))
    return np.array(vecs).T



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Counter, defaultdict
from functools import reduce
import itertools
import random
import warnings

import numpy as np
from scipy.optimize import minimize as scipy_minimizer
from .circuit import Circuit
from .utils import to_inttuple


class AnsatzBase:
    def __init__(self, hamiltonian, n_params):
        self.hamiltonian = hamiltonian
        self.n_params = n_params
        self.n_qubits = self.hamiltonian.max_n() + 1
        self.sparse = None

    def make_sparse(self, fmt='csc', make_method=None):
        """Make sparse matrix. This method may be changed in the future release."""
        if make_method:
            self.sparse = make_method(self.hamiltonian)
        else:
            self.sparse = self.hamiltonian.to_matrix(sparse=fmt)

    def get_circuit(self, params):
        """Make a circuit from parameters."""
        raise NotImplementedError

    def get_energy(self, circuit, sampler):
        """Calculate energy from circuit and sampler."""
        val = 0.0
        for meas in self.hamiltonian:
            c = circuit.copy()
            for op in meas.ops:
                if op.op == "X":
                    c.h[op.n]
                elif op.op == "Y":
                    c.rx(-np.pi / 2)[op.n]
            measured = sampler(c, meas.n_iter())
            for bits, prob in measured.items():
                if sum(bits) % 2:
                    val -= prob * meas.coeff
                else:
                    val += prob * meas.coeff
        return val.real

    def get_energy_sparse(self, circuit):
        """Get energy using sparse matrix. This method may be changed in the future release."""
        return sparse_expectation(self.sparse, circuit.run())

    def get_objective(self, sampler=None):
        """Get an objective function to be optimized."""
        def objective(params):
            circuit = self.get_circuit(params)
            circuit.make_cache()
            return self.get_energy(circuit, sampler)

        def obj_expect(params):
            circuit = self.get_circuit(params)
            circuit.make_cache()
            return self.get_energy_sparse(circuit)

        if sampler is not None:
            return objective
        if self.sparse is None:
            self.make_sparse()
        return obj_expect


class QaoaAnsatz(AnsatzBase):
    """Ansatz for QAOA."""
    def __init__(self, hamiltonian, step=1, init_circuit=None, mixer=None):
        """
        Args:
            hamiltonian (Pauli expr): Hamiltonian for optimization.
            step (optional, int): The number of step for repeat the circuit.
                Generally, larger is precise and slow.
            init_circuit (optional, Circuit): Initial circuit for QAOA.
                By default, `Circuit().h[:]` is used.
                If mixer is specified, this argument is required.
            mixer (optional, list of Pauli expr): Mixer for Quantum Alternative Operator Ansatz.
                By default, mixer is None and normal Quantum Approximate Optimization Algorithm is used.
                This feature is experimental. It may be modified.
        """
        super().__init__(hamiltonian, step * 2)
        self.hamiltonian = hamiltonian.to_expr().simplify()
        if not self.check_hamiltonian():
            raise ValueError("Hamiltonian terms are not commutable")

        self.step = step
        self.n_qubits = self.hamiltonian.max_n() + 1
        if init_circuit:
            self.init_circuit = init_circuit
            if init_circuit.n_qubits > self.n_qubits:
                self.n_qubits = init_circuit.n_qubits
        else:
            if mixer:
                raise ValueError(
                    'init_circuit is required when mixer is not default.')
            self.init_circuit = Circuit(self.n_qubits).h[:]
        self.mixer = mixer
        self.init_circuit.make_cache()
        self.time_evolutions = [
            term.get_time_evolution() for term in self.hamiltonian
        ]
        self.mixer_time_evolutions = [
            term.get_time_evolution() for term in self.mixer
        ] if mixer else []

    def check_hamiltonian(self):
        """Check hamiltonian is commutable. This condition is required for QaoaAnsatz"""
        return self.hamiltonian.is_all_terms_commutable()

    def get_circuit(self, params):
        c = self.init_circuit.copy()
        betas = params[:self.step]
        gammas = params[self.step:]
        for beta, gamma in zip(betas, gammas):
            beta *= np.pi
            gamma *= 2 * np.pi
            for evo in self.time_evolutions:
                evo(c, gamma)
            if self.mixer is None:
                c.rx(beta)[:]
            else:
                for evo in self.mixer_time_evolutions:
                    evo(c, beta)
        return c


class VqeResult:
    def __init__(self, vqe=None, params=None, circuit=None):
        self.vqe = vqe
        self.params = params
        self.circuit = circuit
        self._probs = None

    def most_common(self, n=1):
        return tuple(
            sorted(self.get_probs().items(), key=lambda item: -item[1]))[:n]

    @property
    def probs(self):
        """Get probabilities. This property is obsoleted. Use get_probs()."""
        warnings.warn(
            "VqeResult.probs is obsoleted. " + "Use VqeResult.get_probs().",
            DeprecationWarning)
        return self.get_probs()

    def get_probs(self, sampler=None, rerun=None, store=True):
        """Get probabilities."""
        if rerun is None:
            rerun = sampler is not None
        if self._probs is not None and not rerun:
            return self._probs
        if sampler is None:
            sampler = self.vqe.sampler

        if sampler is None:
            probs = expect(self.circuit.run(returns="statevector"),
                           range(self.circuit.n_qubits))
        else:
            probs = sampler(self.circuit, range(self.circuit.n_qubits))
        if store:
            self._probs = probs
        return probs


class Vqe:
    def __init__(self, ansatz, minimizer=None, sampler=None):
        self.ansatz = ansatz
        self.minimizer = minimizer or get_scipy_minimizer(method="Powell",
                                                          options={
                                                              "ftol": 5.0e-2,
                                                              "xtol": 5.0e-2,
                                                              "maxiter": 1000
                                                          })
        self.sampler = sampler

    def run(self, verbose=False):
        objective = self.ansatz.get_objective(self.sampler)
        if verbose:

            def verbose_objective(objective):
                def f(params):
                    val = objective(params)
                    print("params:", params, "val:", val)
                    return val

                return f

            objective = verbose_objective(objective)
        params = self.minimizer(objective, self.ansatz.n_params)
        c = self.ansatz.get_circuit(params)
        return VqeResult(self, params, c)

    @property
    def result(self):
        """Vqe.result is deprecated. Use `result = Vqe.run()`."""
        warnings.warn("Vqe.result is deprecated. Use `result = Vqe.run()`",
                      DeprecationWarning)
        return self._result if self._result is not None else VqeResult()


def get_scipy_minimizer(**kwargs):
    """Get minimizer which uses `scipy.optimize.minimize`"""
    def minimizer(objective, n_params):
        params = [random.random() for _ in range(n_params)]
        result = scipy_minimizer(objective, params, **kwargs)
        return result.x

    return minimizer


def expect(qubits, meas):
    "For the VQE simulation without sampling."
    result = {}
    i = np.arange(len(qubits))
    meas = tuple(meas)

    def to_mask(n):
        return reduce(
            lambda acc, im: acc | (n & (1 << im[0])) << (im[1] - im[0]),
            enumerate(meas), 0)

    def to_key(k):
        return tuple(1 if k & (1 << i) else 0 for i in meas)

    mask = reduce(lambda acc, v: acc | (1 << v), meas, 0)

    cnt = defaultdict(float)
    for i, v in enumerate(qubits):
        p = v.real**2 + v.imag**2
        if p != 0.0:
            cnt[i & mask] += p
    return {to_key(k): v for k, v in cnt.items()}


def non_sampling_sampler(circuit, meas):
    """Calculate the expectations without sampling."""
    meas = tuple(meas)
    n_qubits = circuit.n_qubits
    return expect(circuit.run(returns="statevector"), meas)


def get_measurement_sampler(n_sample, run_options=None):
    """Returns a function which get the expectations by sampling the measured circuit"""
    if run_options is None:
        run_options = {}

    def sampling_by_measurement(circuit, meas):
        def reduce_bits(bits, meas):
            bits = [int(x) for x in bits[::-1]]
            return tuple(bits[m] for m in meas)

        meas = tuple(meas)
        circuit.measure[meas]
        counter = circuit.run(shots=n_sample, returns="shots", **run_options)
        counts = Counter(
            {reduce_bits(bits, meas): val
             for bits, val in counter.items()})
        return {k: v / n_sample for k, v in counts.items()}

    return sampling_by_measurement


def get_state_vector_sampler(n_sample):
    """Returns a function which get the expectations by sampling the state vector"""
    def sampling_by_measurement(circuit, meas):
        val = 0.0
        e = expect(circuit.run(returns="statevector"), meas)
        bits, probs = zip(*e.items())
        dists = np.random.multinomial(n_sample, probs) / n_sample
        return dict(zip(tuple(bits), dists))

    return sampling_by_measurement


def get_qiskit_sampler(backend, **execute_kwargs):
    """Returns a function which get the expectation by sampling via Qiskit.

    This function requires `qiskit` module.
    """
    try:
        import qiskit
    except ImportError:
        raise ImportError(
            "blueqat.vqe.get_qiskit_sampler() requires qiskit. Please install before call this function."
        )
    try:
        shots = execute_kwargs['shots']
    except KeyError:
        execute_kwargs['shots'] = shots = 1024

    def reduce_bits(bits, meas):
        # In Qiskit 0.6.1, For example
        # Aer backend returns bit string and IBMQ backend returns hex string.
        # Sample code:
        """
from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister, execute, Aer, IBMQ
IBMQ.load_accounts()
q, c = QuantumRegister(4, 'q'), ClassicalRegister(4, 'c')
circ = QuantumCircuit(q, c)
circ.x(q[1])
for i in range(4):
    circ.measure(q[i], c[i])
print("Aer qasm_simulator_py")
print(execute(circ, Aer.get_backend('qasm_simulator_py')).result().get_counts())
print("IBMQ ibmq_qasm_simulator")
print(execute(circ, IBMQ.get_backend('ibmq_qasm_simulator')).result().get_counts())
        """
        # The result is,
        # Aer: {'0010': 1024}
        # IBMQ: {'0x2': 1024}
        # This is workaround for this IBM's specifications.
        if bits.startswith("0x"):
            bits = int(bits, base=16)
            bits = "0" * 100 + format(bits, "b")
        bits = [int(x) for x in bits[::-1]]
        return tuple(bits[m] for m in meas)

    def sampling(circuit, meas):
        meas = tuple(meas)
        if not meas:
            return {}
        circuit.measure[meas]
        result = circuit.run_with_ibmq(qiskit_backend=backend,
                                       returns="qiskit_result",
                                       **execute_kwargs)
        counts = Counter({
            reduce_bits(bits, meas): val
            for bits, val in result.get_counts().items()
        })
        return {k: v / shots for k, v in counts.items()}

    return sampling


def sparse_expectation(mat, vec):
    """Calculate expectation value <vec|mat|vec>.

    Args:
        mat (scipy sparse matrix): Sparse matrix
        vec (numpy array): Vector

    Returns:
        (Real part of) expectation value <vec|mat|vec>.
        Remarks: when mat is Hermitian, <vec|mat|vec> is real.
    """
    return np.vdot(vec, mat.dot(vec)).real



from typing import Callable, Dict, List, Optional, Sequence, Tuple, Type, Union
import warnings

import numpy as np

from ..circuit import Circuit
from ..gate import *
from .backendbase import Backend

_eye = np.eye(2, dtype=complex)

BASIS_TABLE: Dict[str, Type[TwoQubitGate]] = {
    'cx': CXGate,
    'cz': CZGate,
    'zz': ZZGate,
}

DECOMPOSE_TABLE: Dict[str, Dict[str, Tuple[
    np.ndarray, np.ndarray, tuple, np.ndarray, np.ndarray]]] = {
        'cx': {
            'cz': (_eye.copy(),
                   np.array([[1, 1], [1, -1]], dtype=complex) / np.sqrt(2), (),
                   _eye.copy(),
                   np.array([[1, 1], [1, -1]], dtype=complex) / np.sqrt(2)),
            'zz':
            (_eye.copy(), np.array([[1j, -1j], [1j, 1j]]) / np.sqrt(2), (),
             np.diag([1 - 1j, 1 + 1j]) / np.sqrt(2),
             np.array([[0.5 - 0.5j, 0.5 - 0.5j], [-0.5 - 0.5j, 0.5 + 0.5j]])),
        },
        'cz': {
            'cx': (_eye.copy(),
                   np.array([[1, 1], [1, -1]], dtype=complex) / np.sqrt(2), (),
                   _eye.copy(),
                   np.array([[1, 1], [1, -1]], dtype=complex) / np.sqrt(2)),
            'zz': (_eye.copy(), np.array([[1j, 0], [0, -1j]]), (),
                   np.diag([1 - 1j, 1 + 1j]) / np.sqrt(2),
                   np.array([[1 - 1j, 0], [0, -1 - 1j]]) / np.sqrt(2)),
        },
        'zz': {
            'cx':
            (_eye.copy(), np.array([[-1j, -1j], [1j, -1j]]) / np.sqrt(2), (),
             np.diag([1 + 1j, 1 - 1j]) / np.sqrt(2),
             np.array([[0.5 + 0.5j, -0.5 + 0.5j], [0.5 + 0.5j, 0.5 - 0.5j]])),
            'cz': (_eye.copy(), np.array([[-1j, 0], [0, 1j]]), (),
                   np.diag([1 + 1j, 1 - 1j]) / np.sqrt(2),
                   np.array([[1 + 1j, 0], [0, -1 + 1j]]) / np.sqrt(2)),
        },
    }


class TwoQubitGateDecomposingTranspiler(Backend):
    """Decomposite two qubit gate."""
    @staticmethod
    def _run_inner(
            gates: List[Operation], operations: List[Operation],
            singlemats: List[np.ndarray], n_qubits: int, basis: Sequence[str],
            mat1_decomposer: Callable[[OneQubitGate],
                                      List[Operation]]) -> None:
        basisgate = BASIS_TABLE[basis[0]]
        table = DECOMPOSE_TABLE[basis[0]]
        for gate in gates:
            if not isinstance(gate, Gate):
                # Non-gate operations.
                for t in gate.target_iter(n_qubits):
                    if not np.allclose(singlemats[t], _eye):
                        operations += mat1_decomposer(
                            Mat1Gate((t, ), singlemats[t]))
                        singlemats[t] = _eye.copy()
                operations.append(gate)
            elif gate.n_qargs == 1:
                for t in gate.target_iter(n_qubits):
                    np.matmul(gate.matrix(), singlemats[t], out=singlemats[t])
            elif gate.lowername in basis:
                for t1, t2 in gate.control_target_iter(n_qubits):
                    if not np.allclose(singlemats[t1], _eye):
                        operations += mat1_decomposer(
                            Mat1Gate((t1, ), singlemats[t1]))
                        singlemats[t1] = _eye.copy()
                    if not np.allclose(singlemats[t2], _eye):
                        operations += mat1_decomposer(
                            Mat1Gate((t2, ), singlemats[t2]))
                        singlemats[t2] = _eye.copy()
                operations.append(gate)
            elif gate.lowername in table:
                for t1, t2 in gate.control_target_iter(n_qubits):
                    l1, l2, gparams, r1, r2 = table[gate.lowername]
                    np.matmul(l1, singlemats[t1], out=singlemats[t1])
                    np.matmul(l2, singlemats[t2], out=singlemats[t2])
                    if not np.allclose(singlemats[t1], _eye):
                        operations += mat1_decomposer(
                            Mat1Gate((t1, ), singlemats[t1]))
                    singlemats[t1] = r1.copy()
                    if not np.allclose(singlemats[t2], _eye):
                        operations += mat1_decomposer(
                            Mat1Gate((t2, ), singlemats[t2]))
                    singlemats[t2] = r2.copy()
                    operations.append(basisgate((t1, t2), *gparams))
            else:
                TwoQubitGateDecomposingTranspiler._run_inner(
                    gate.fallback(n_qubits), operations, singlemats, n_qubits,
                    basis, mat1_decomposer)

    @staticmethod
    def run(gates: List[Operation],
            n_qubits: int,
            *,
            basis: Union[str, Sequence[str]],
            mat1_decomposer: Optional[Callable[[OneQubitGate],
                                               List[Operation]]] = None,
            **kwargs) -> Circuit:
        """
        Args:
            basis (str or non-empty tuple of str): Name (lowername) of basis 2 qubits gate.
                If multiple basis gates are given, given basis gates are not decomposited and
                non-basis gate is decomposited to first available basis gate.
        """
        if kwargs:
            warnings.warn(f"Unknown argument {kwargs}")
        if mat1_decomposer is None:
            mat1_decomposer = lambda g: [g]

        if not basis:
            raise ValueError('No basis provided.')
        if isinstance(basis, str):
            basis = (basis, )
        if basis[0] not in DECOMPOSE_TABLE:
            raise ValueError('Unsupported basis.')
        operations = []
        singlemats = [_eye.copy() for _ in range(n_qubits)]
        TwoQubitGateDecomposingTranspiler._run_inner(gates, operations,
                                                     singlemats, n_qubits,
                                                     basis, mat1_decomposer)
        for i, mat in enumerate(singlemats):
            if not np.allclose(mat, _eye):
                operations += mat1_decomposer(Mat1Gate((i, ), mat))
        return Circuit(n_qubits, operations)


def two_qubit_gate_decompose(
    circuit: Circuit,
    basis: Union[str, Sequence[str]],
    mat1_decomposer: Optional[Callable[[OneQubitGate], List[Operation]]] = None
) -> Circuit:
    """Function for decompose two qubit gates.

    This function is experimenatal feature.
    It may be deleted or destructively changed in future release."""
    return TwoQubitGateDecomposingTranspiler.run(
        circuit.ops,
        circuit.n_qubits,
        basis=basis,
        mat1_decomposer=mat1_decomposer)



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
`gate` module implements quantum gate operations.
This module is internally used.
"""

import copy
from typing import Any, Callable, Dict, List, Optional, Tuple

from ..gate import Operation


class Backend:
    """Abstract quantum gate processor backend class."""
    def copy(self) -> "Backend":
        """Returns (deep)copy of Backend.

        Backend developer must support `copy` method.
        If required, the developer can override this method.
        """
        return copy.deepcopy(self)

    def _preprocess_run(self, gates: List[Operation], n_qubits: int,
                        args: Tuple[Any], kwargs: Dict[Any, Any]) -> Any:
        """Preprocess of backend run.
        Backend developer can override this function.
        """
        return gates, None

    def _postprocess_run(self, ctx: Any) -> Any:
        """Postprocess of backend run
        Backend developer can override this function.
        """
        return None

    def _run_gates(self, gates: List[Operation], n_qubits: int,
                   ctx: Any) -> Any:
        """Iterate gates and call backend's action for each gates"""
        for gate in gates:
            action = self._get_action(gate)
            if action is not None:
                ctx = action(gate, ctx)
            else:
                ctx = self._run_gates(gate.fallback(n_qubits), n_qubits, ctx)
        return ctx

    def _run(self, gates: List[Operation], n_qubits: int, args: Tuple[Any],
             kwargs: Dict[Any, Any]) -> Any:
        """Default implementation of `Backend.run`.
        Backend developer shouldn't override this function, but override `run` instead of this.

        The default flow of running is:
            1. preprocessing
            2. call the gate action which defined in backend
            3. postprocessing

        Backend developer can:
            1. Define preprocessing process. Override `_preprocess_run`
            2. Define the gate action. Define methods `gate_{gate.lowername}`,
               for example, `gate_x` for X gate, `gate_cx` for CX gate.
            3. Define postprocessing process (and make return value). Override `_postprocess_run`
        Otherwise, the developer can override `run` method if they want to change the flow of run.
        """
        gates, ctx = self._preprocess_run(gates, n_qubits, args, kwargs)
        self._run_gates(gates, n_qubits, ctx)
        return self._postprocess_run(ctx)

    def make_cache(self, gates: List[Operation], n_qubits: int):
        """Make internal cache to reduce the time of running the circuit.

        Some backends may implement this method. Otherwise, this method do nothing.
        This is a temporary API and may changed or deprecated in future."""
        return None

    def run(self, gates: List[Operation], n_qubits: int, *args, **kwargs):
        """Run the backend."""
        return self._run(gates, n_qubits, args, kwargs)

    def _get_action(self, gate: Operation) -> Optional[Callable]:
        try:
            return getattr(self, "gate_" + gate.lowername)
        except AttributeError:
            return None

    def _has_action(self, gate: Operation) -> bool:
        return hasattr(self, "gate_" + gate.lowername)

    def _resolve_fallback(self, gates: Operation,
                          n_qubits: int) -> List[Operation]:
        """Resolve fallbacks and flatten gates."""
        flattened = []
        for g in gates:
            if self._has_action(g):
                flattened.append(g)
            else:
                flattened += self._resolve_fallback(g.fallback(n_qubits),
                                                    n_qubits)
        return flattened



from .. import Circuit
from .. import gate as g

def flatten(c: Circuit) -> Circuit:
    """expands slice and multiple targets into single target"""
    n_qubits = c.n_qubits
    ops = []
    for op in c.ops:
        if isinstance(op, (g.OneQubitGate, g.Measurement, g.Reset)):
            ops += [op.create(t, op.params) for t in op.target_iter(n_qubits)]
        elif isinstance(op, g.TwoQubitGate):
            ops += [op.create(t, op.params) for t in op.control_target_iter(n_qubits)]
        else:
            raise ValueError(f"Cannot process operation {op.lowername}.")
    return Circuit(n_qubits, ops)



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..gate import *
from .backendbase import Backend
from .qasm_output_backend import QasmOutputBackend


class QasmParsableBackend(Backend):
    """Backend for third-party library which can read OpenQASM """
    def __init__(self, qasm_runner):
        """Specify qasm_runner which receive OpenQASM text and returns result.

        Args:
            qasm_runner (function (qasm: str, *args, **kwargs) -> Result):
                An function which receives OpenQASM and some arguments and returns a result.
        """
        self.to_qasm = QasmOutputBackend()
        self.qasm_runner = qasm_runner

    def run(self, gates, n_qubits, *args, **kwargs):
        qasm = self.to_qasm.run(gates, n_qubits)
        return self.qasm_runner(qasm, *args, **kwargs)


def generate_backend(qasm_runner):
    """Generate a wrapper of QasmParsableBackend from qasm_runner.

    Due to Blueqat's backend specifications, normally, cannot give arguments to
    QasmParsableBackend's constructor.
    This function wrap the class for specify `qasm_runner` to QasmParsableBackend's constructor.

    Args:
        qasm_runner (function (qasm: str, *args, **kwargs) -> Result):
            An function which receives OpenQASM and some arguments and returns a result.

    Returns:
        function () -> QasmParsableBackend(qasm_runner)
    """
    return lambda: QasmParsableBackend(qasm_runner)



"""
`gate` module implements quantum gate operations.
This module is internally used.
"""

import cmath
import math
from typing import Callable, Iterable, Iterator, List, NoReturn, Tuple, Type, TypeVar, Union

import numpy as np

from .typing import Targets

_Op = TypeVar('_Op', bound='Operation')


class Operation:
    """Abstract quantum circuit operation class."""

    lowername: str = ''
    """Lower name of the operation."""
    @property
    def uppername(self) -> str:
        """Upper name of the operation."""
        return self.lowername.upper()

    def __init__(self, targets: Targets, params=(), **kwargs) -> None:
        if self.lowername == '':
            raise ValueError(
                f"{self.__class__.__name__}.lowername is not defined.")
        self.params = params
        self.kwargs = kwargs
        self.targets = targets

    def target_iter(self, n_qubits: int) -> Iterator[int]:
        """The generator which yields the target qubits."""
        return slicing(self.targets, n_qubits)

    @classmethod
    def create(cls: Type[_Op], targets: Targets, params: tuple) -> _Op:
        """Create an operation."""
        raise NotImplementedError(f"{cls.__name__}.create() is not defined.")

    def fallback(self, n_qubits: int) -> List['Operation']:
        """Returns alternative operations to make equivalent circuit."""
        raise NotImplementedError(
            f"The fallback of {self.__class__.__name__} operation is not defined."
        )

    def dagger(self) -> 'Operation':
        """Returns the Hermitian conjugate of `self`."""
        raise NotImplementedError(
            f"Hermitian conjugate of {self.__class__.__name__} operation is not provided."
        )

    def _str_args(self) -> str:
        """Returns printable string of args."""
        if not self.params:
            return ''
        return '(' + ', '.join(str(param) for param in self.params) + ')'

    def _str_targets(self) -> str:
        """Returns printable string of targets."""
        def _slice_to_str(obj):
            if isinstance(obj, slice):
                start = '' if obj.start is None else str(obj.start.__index__())
                stop = '' if obj.stop is None else str(obj.stop.__index__())
                if obj.step is None:
                    return f'{start}:{stop}'
                step = str(obj.step.__index__())
                return f'{start}:{stop}:{step}'
            return str(obj.__index__())

        if isinstance(self.targets, tuple):
            return f"[{', '.join(_slice_to_str(target) for target in self.targets)}]"
        return f"[{_slice_to_str(self.targets)}]"

    def __str__(self) -> str:
        str_args = self._str_args()
        str_targets = self._str_targets()
        return f'{self.lowername}{str_args}{str_targets}'


class Gate(Operation):
    """Abstract quantum gate class."""
    @property
    def n_qargs(self) -> int:
        """Number of qubit arguments of this gate."""
        raise NotImplementedError()

    def fallback(self, n_qubits: int) -> List['Gate']:
        """Returns alternative gates to make equivalent circuit."""
        raise NotImplementedError(
            f"The fallback of {self.__class__.__name__} gate is not defined.")

    def dagger(self) -> 'Gate':
        """Returns the Hermitian conjugate of `self`."""
        raise NotImplementedError(
            "Hermitian conjugate of this gate is not provided.")

    def matrix(self) -> np.ndarray:
        """Returns the matrix of implementations.

        (Non-abstract) subclasses of Gate must implement this method.
        WARNING: qubit order specifications of multi qubit gate is still not defined.
        """
        raise NotImplementedError()


class OneQubitGate(Gate):
    """Abstract quantum gate class for 1 qubit gate."""

    u_params: Tuple[float, float, float, float]
    """Params for U gate."""
    @property
    def n_qargs(self) -> int:
        return 1

    def _make_fallback_for_target_iter(
            self, n_qubits: int,
            fallback: Callable[[int], List['Gate']]) -> List['Gate']:
        gates = []
        for t in self.target_iter(n_qubits):
            gates += fallback(t)
        return gates

    def fallback(self, n_qubits: int) -> List['Gate']:
        if self.u_params:
            return [UGate(self.targets, *self.u_params, **self.kwargs)]
        try:
            mat = self.matrix()
        except NotImplementedError:
            return super().fallback(n_qubits)
        return [Mat1Gate(self.targets, mat, **self.kwargs)]


class TwoQubitGate(Gate):
    """Abstract quantum gate class for 2 qubits gate."""

    cu_params: Tuple[float, float, float, float]

    @property
    def n_qargs(self):
        return 2

    def control_target_iter(self, n_qubits: int) -> Iterator[Tuple[int, int]]:
        """The generator which yields the tuples of (control, target) qubits."""
        return qubit_pairs(self.targets, n_qubits)

    def _make_fallback_for_control_target_iter(
            self, n_qubits: int,
            fallback: Callable[[int, int], List['Gate']]) -> List['Gate']:
        gates = []
        for c, t in self.control_target_iter(n_qubits):
            gates += fallback(c, t)
        return gates

    def fallback(self, n_qubits: int) -> List['Gate']:
        if self.cu_params:
            return [CUGate(self.targets, *self.cu_params, **self.kwargs)]
        return super().fallback(n_qubits)


class HGate(OneQubitGate):
    """Hadamard gate"""
    lowername = "h"
    u_params = (-math.pi / 2.0, math.pi, 0.0, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'HGate':
        return cls(targets, *params)

    def dagger(self):
        return self

    def matrix(self):
        return np.array([[1, 1], [1, -1]], dtype=complex) / math.sqrt(2)


class IGate(OneQubitGate):
    """Identity gate"""
    lowername = "i"
    u_params = (0.0, 0.0, 0.0, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'IGate':
        return cls(targets, *params)

    def fallback(self, _):
        return []

    def dagger(self):
        return self

    def matrix(self):
        return np.eye(2)


class Mat1Gate(OneQubitGate):
    """Arbitrary 2x2 matrix gate

    `mat` is expected a 2x2 unitary matrix, but not checked.
    (If unexpected matrix is given, backend may raises error or returns weird result)
    """
    lowername = "mat1"
    u_params = None

    def __init__(self, targets, mat: np.ndarray, **kwargs):
        super().__init__(targets, (mat, ), **kwargs)
        self.mat = mat

    @classmethod
    def create(cls, targets, params) -> 'Mat1Gate':
        return cls(targets, params[0])

    def dagger(self):
        return Mat1Gate(self.targets, self.mat.T.conjugate(), **self.kwargs)

    def matrix(self):
        return self.mat

    def fallback(self, n_qubits: int) -> List['Gate']:
        raise NotImplementedError(
            'Fallback implementation for Mat1Gate is not available.')


class PhaseGate(OneQubitGate):
    """Rotate-Z gate but global phase is different.

    Global phase doesn't makes any difference of measured result.
    You may use RZ gate or U1 gate instead, but distinguishing these gates
    may better for debugging or future improvement.

    furthermore, phase gate may efficient for simulating.
    (It depends on backend implementation.
    But matrix of phase gate is simpler than RZ gate or U1 gate.)"""
    lowername = "phase"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta, ), **kwargs)
        self.theta = theta
        self.u_params = (0.0, self.theta, 0.0, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'PhaseGate':
        return cls(targets, params[0])

    def dagger(self):
        return PhaseGate(self.targets, -self.theta, **self.kwargs)

    def matrix(self):
        return np.array([[1, 0], [0, cmath.exp(1j * self.theta)]],
                        dtype=complex)


class RXGate(OneQubitGate):
    """Rotate-X gate"""
    lowername = "rx"

    def __init__(self, targets, theta: float, **kwargs):
        super().__init__(targets, (theta, ), **kwargs)
        self.theta = theta
        self.u_params = (theta, -math.pi / 2.0, math.pi / 2.0, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'RXGate':
        return cls(targets, params[0])

    def dagger(self):
        return RXGate(self.targets, -self.theta, **self.kwargs)

    def matrix(self):
        t = self.theta * 0.5
        a = math.cos(t)
        b = -1j * math.sin(t)
        return np.array([[a, b], [b, a]], dtype=complex)


class RYGate(OneQubitGate):
    """Rotate-Y gate"""
    lowername = "ry"

    def __init__(self, targets, theta: float, **kwargs):
        super().__init__(targets, (theta, ), **kwargs)
        self.theta = theta
        self.u_params = (theta, 0.0, 0.0, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'RYGate':
        return cls(targets, params[0])

    def dagger(self):
        return RYGate(self.targets, -self.theta, **self.kwargs)

    def matrix(self):
        t = self.theta * 0.5
        a = math.cos(t)
        b = math.sin(t)
        return np.array([[a, -b], [b, a]], dtype=complex)


class RZGate(OneQubitGate):
    """Rotate-Z gate"""
    lowername = "rz"

    def __init__(self, targets, theta: float, **kwargs):
        super().__init__(targets, (theta, ), **kwargs)
        self.theta = theta
        self.u_params = (0.0, 0.0, theta, -0.5 * theta)

    @classmethod
    def create(cls, targets, params) -> 'RZGate':
        return cls(targets, params[0])

    def dagger(self):
        return RZGate(self.targets, -self.theta, **self.kwargs)

    def matrix(self):
        a = cmath.exp(0.5j * self.theta)
        return np.array([[a.conjugate(), 0], [0, a]], dtype=complex)


class SGate(OneQubitGate):
    """S gate"""
    lowername = "s"
    u_params = (0.0, 0.0, math.pi / 2, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'SGate':
        return cls(targets, params)

    def dagger(self):
        return SDagGate(self.targets, self.params, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_target_iter(
            n_qubits, lambda t: [PhaseGate(t, math.pi / 2)])

    def matrix(self):
        return np.array([[1, 0], [0, 1j]])


class SDagGate(OneQubitGate):
    """Dagger of S gate"""
    lowername = "sdg"
    u_params = (0.0, 0.0, -math.pi / 2, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'SDagGate':
        return cls(targets, params)

    def dagger(self):
        return SGate(self.targets, self.params, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_target_iter(
            n_qubits, lambda t: [PhaseGate(t, -math.pi / 2)])

    def matrix(self):
        return np.array([[1, 0], [0, -1j]])


class SXGate(OneQubitGate):
    """sqrt(X) gate

    This is equivalent as RX(π/2) * (1 + i) / √2."""
    lowername = "sx"
    u_params = (math.pi / 2.0, -math.pi / 2.0, math.pi / 2.0, math.pi / 4.0)

    @classmethod
    def create(cls, targets, params) -> 'SXGate':
        return cls(targets, params)

    def dagger(self):
        return SXDagGate(self.targets, self.params, **self.kwargs)

    def matrix(self):
        return np.array([[1 + 1j, 1 - 1j], [1 - 1j, 1 + 1j]])


class SXDagGate(OneQubitGate):
    """sqrt(X)† gate"""
    lowername = "sxdg"
    u_params = (-math.pi / 2.0, -math.pi / 2.0, math.pi / 2.0, -math.pi / 4.0)

    def dagger(self):
        return SXGate(self.targets, self.params, **self.kwargs)

    def matrix(self):
        return np.array([[1 - 1j, 1 + 1j], [1 + 1j, 1 - 1j]])


class TGate(OneQubitGate):
    """T ($\\pi/8$) gate"""
    lowername = "t"
    u_params = (0, 0, math.pi / 4.0, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'TGate':
        return cls(targets, params)

    def dagger(self):
        return TDagGate(self.targets, self.params, **self.kwargs)

    def fallback(self, n_qubits):
        return [PhaseGate(self.targets, math.pi / 4, **self.kwargs)]

    def matrix(self):
        return np.array([[1, 0], [0, math.exp(math.pi * 0.25)]])


class TDagGate(OneQubitGate):
    """Dagger of T ($\\pi/8$) gate"""
    lowername = "tdg"
    u_params = (0, 0, -math.pi / 4.0, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'TDagGate':
        return cls(targets, params)

    def dagger(self):
        return TGate(self.targets, self.params, **self.kwargs)

    def fallback(self, n_qubits):
        return [PhaseGate(self.targets, -math.pi / 4, **self.kwargs)]

    def matrix(self):
        return np.array([[1, 0], [0, math.exp(math.pi * -0.25)]])


class ToffoliGate(Gate):
    """Toffoli (CCX) gate"""
    lowername = "ccx"

    @property
    def n_qargs(self):
        return 3

    @classmethod
    def create(cls, targets, params) -> 'ToffoliGate':
        return cls(targets, params)

    def dagger(self):
        return self

    def fallback(self, n_qubits):
        c1, c2, t = self.targets
        return [
            HGate(t),
            CCZGate((c1, c2, t)),
            HGate(t),
        ]

    def matrix(self):
        return np.array([[1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0],
                         [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0, 0, 1],
                         [0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0],
                         [0, 0, 0, 0, 0, 0, 1, 0], [0, 0, 0, 1, 0, 0, 0, 0]],
                        dtype=complex)


class UGate(OneQubitGate):
    """Arbitrary 1 qubit unitary gate including global phase.

    U(θ, φ, λ, γ = 0.0) = e^iγ * array([
        [cos(θ/2), -e^iλ sin(θ/2)],
        [e^iφ sin(θ/2), e^i(φ+λ) cos(θ/2)]])

    Note: If SU matrix is required, U(θ, φ, λ, (φ + λ) / 2) works fine.
    """
    lowername = "u"

    def __init__(self,
                 targets,
                 theta: float,
                 phi: float,
                 lam: float,
                 gamma: float = 0.0,
                 **kwargs):
        super().__init__(targets, (theta, phi, lam, gamma), **kwargs)
        self.theta = theta
        self.phi = phi
        self.lam = lam
        self.gamma = gamma
        self.u_params = (theta, phi, lam, gamma)

    @classmethod
    def create(cls, targets, params) -> 'UGate':
        return cls(targets, *params)

    def dagger(self):
        return UGate(self.targets, -self.theta, -self.lam, -self.phi,
                     -self.gamma, **self.kwargs)

    def fallback(self, n_qubits: int) -> List['Gate']:
        raise NotImplementedError(
            'Fallback implementation for UGate is not available.')

    def matrix(self):
        t, p, l, g = self.params
        gphase = cmath.exp(1j * g)
        cos_t = math.cos(0.5 * t)
        sin_t = math.sin(0.5 * t)
        return np.array(
            [[cos_t, -cmath.exp(1j * l) * sin_t],
             [cmath.exp(1j * p) * sin_t,
              cmath.exp(1j * (p + l)) * cos_t]],
            dtype=complex) * gphase


class XGate(OneQubitGate):
    """Pauli's X gate"""
    lowername = "x"
    u_params = (math.pi, 0.0, math.pi, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'XGate':
        return cls(targets, params)

    def dagger(self):
        return self

    def matrix(self):
        return np.array([[0, 1], [1, 0]], dtype=complex)


class YGate(OneQubitGate):
    """Pauli's Y gate"""
    lowername = "y"
    u_params = (math.pi, math.pi / 2, math.pi / 2, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'YGate':
        return cls(targets, params)

    def dagger(self):
        return self

    def matrix(self):
        return np.array([[0, -1j], [1j, 0]])


class ZGate(OneQubitGate):
    """Pauli's Z gate"""
    lowername = "z"
    u_params = (0.0, 0.0, math.pi, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'ZGate':
        return cls(targets, params)

    def dagger(self):
        return self

    def matrix(self):
        return np.array([[1, 0], [0, -1]], dtype=complex)


class CCZGate(Gate):
    """2-Controlled Z gate"""
    lowername = "ccz"

    @property
    def n_qargs(self):
        return 3

    @classmethod
    def create(cls, targets, params) -> 'CCZGate':
        return cls(targets, params)

    def fallback(self, n_qubits):
        c1, c2, t = self.targets
        return [
            CXGate((c2, t)),
            TDagGate(t),
            CXGate((c1, t)),
            TGate(t),
            CXGate((c2, t)),
            TDagGate(t),
            CXGate((c1, t)),
            TGate(c2),
            TGate(t),
            CXGate((c1, c2)),
            TGate(c1),
            TDagGate(c2),
            CXGate((c1, c2)),
        ]

    def dagger(self):
        return self

    def matrix(self):
        return np.array([[1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0],
                         [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0, 0],
                         [0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0],
                         [0, 0, 0, 0, 0, 0, 1, 0], [0, 0, 0, 0, 0, 0, 0, -1]],
                        dtype=complex)


class CHGate(TwoQubitGate):
    """Controlled-H gate"""
    lowername = "ch"
    cu_params = (-math.pi / 2.0, math.pi, 0.0, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'CHGate':
        return cls(targets, params)

    def dagger(self):
        return self

    def matrix(self):
        a = 1.0 / math.sqrt(2)
        return np.array(
            [[1, 0, 0, 0], [0, a, 0, a], [0, 0, 1, 0], [0, a, 0, -a]],
            dtype=complex)


class CPhaseGate(TwoQubitGate):
    """Rotate-Z gate but phase is different."""
    lowername = "cphase"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta, ), **kwargs)
        self.theta = theta
        self.cu_params = (0.0, self.theta, 0.0, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'CPhaseGate':
        return cls(targets, params[0])

    def dagger(self):
        return CPhaseGate(self.targets, -self.theta, **self.kwargs)

    def matrix(self):
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0],
                         [0, 0, 0, cmath.exp(1j * self.theta)]],
                        dtype=complex)


class CRXGate(TwoQubitGate):
    """Rotate-X gate"""
    lowername = "crx"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta, ), **kwargs)
        self.theta = theta
        self.cu_params = (theta, -math.pi / 2.0, math.pi / 2.0, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'CRXGate':
        return cls(targets, params[0])

    def dagger(self):
        return CRXGate(self.targets, -self.theta, **self.kwargs)

    def matrix(self):
        t = self.theta * 0.5
        a = math.cos(t)
        b = -1j * math.sin(t)
        return np.array(
            [[1, 0, 0, 0], [0, a, 0, b], [0, 0, 1, 0], [0, b, 0, a]],
            dtype=complex)


class CRYGate(TwoQubitGate):
    """Rotate-Y gate"""
    lowername = "cry"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta, ), **kwargs)
        self.theta = theta
        self.cu_params = (theta, 0.0, 0.0, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'CRYGate':
        return cls(targets, params[0])

    def dagger(self):
        return CRYGate(self.targets, -self.theta, **self.kwargs)

    def matrix(self):
        t = self.theta * 0.5
        a = math.cos(t)
        b = math.sin(t)
        return np.array(
            [[1, 0, 0, 0], [0, a, 0, -b], [0, 0, 1, 0], [0, b, 0, a]],
            dtype=complex)


class CRZGate(TwoQubitGate):
    """Rotate-Z gate"""
    lowername = "crz"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta, ), **kwargs)
        self.theta = theta
        self.cu_params = (0.0, 0.0, theta, -0.5 * theta)

    @classmethod
    def create(cls, targets, params) -> 'CRZGate':
        return cls(targets, params[0])

    def dagger(self):
        return CRZGate(self.targets, -self.theta, **self.kwargs)

    def matrix(self):
        a = cmath.exp(0.5j * self.theta)
        return np.array([[1, 0, 0, 0], [0, a.conjugate(), 0, 0], [0, 0, 1, 0],
                         [0, 0, 0, a]],
                        dtype=complex)


class CSwapGate(Gate):
    """Controlled SWAP gate"""
    lowername = "cswap"

    @property
    def n_qargs(self):
        return 3

    @classmethod
    def create(cls, targets, params) -> 'CSwapGate':
        return cls(targets, params)

    def dagger(self):
        return self

    def fallback(self, n_qubits):
        c, t1, t2 = self.targets
        return [CXGate((t2, t1)), ToffoliGate((c, t1, t2)), CXGate((t2, t1))]

    def matrix(self):
        return np.array([[1, 0, 0, 0, 0, 0, 0, 0], [0, 1, 0, 0, 0, 0, 0, 0],
                         [0, 0, 1, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0],
                         [0, 0, 0, 0, 1, 0, 0, 0], [0, 0, 0, 1, 0, 0, 0, 0],
                         [0, 0, 0, 0, 0, 0, 1, 0], [0, 0, 0, 0, 0, 0, 0, 1]],
                        dtype=complex)


class CUGate(TwoQubitGate):
    """Controlled-U gate."""
    lowername = "cu"

    def __init__(self,
                 targets,
                 theta: float,
                 phi: float,
                 lam: float,
                 gamma: float = 0.0,
                 **kwargs):
        super().__init__(targets, (theta, phi, lam, gamma), **kwargs)
        self.theta = theta
        self.phi = phi
        self.lam = lam
        self.gamma = gamma

    @classmethod
    def create(cls, targets, params) -> 'CUGate':
        return cls(targets, *params)

    def dagger(self):
        return CUGate(self.targets, -self.theta, -self.lam, -self.phi,
                      -self.gamma, **self.kwargs)

    def fallback(self, n_qubits: int) -> List['Gate']:
        theta, phi, lam, gamma = self.params
        return self._make_fallback_for_control_target_iter(
            n_qubits, lambda c, t: [
                PhaseGate((c, ), gamma + 0.5 * (lam + phi)),
                PhaseGate((t, ), 0.5 * (lam - phi)),
                CXGate((c, t)),
                UGate((t, ), -0.5 * theta, 0.0, -0.5 * (phi + lam)),
                CXGate((c, t)),
                UGate((t, ), 0.5 * theta, phi, 0.0)
            ])

    def matrix(self):
        t, p, l, g = self.params
        cos_t = math.cos(0.5 * t)
        sin_t = math.sin(0.5 * t)
        return np.array([[
            1, 0, 0, 0
        ], [0,
            cmath.exp(1j * g) * cos_t, 0, -cmath.exp(1j * (g + l)) * sin_t],
                         [0, 0, 1, 0],
                         [
                             0,
                             cmath.exp(1j * (g + p)) * sin_t, 0,
                             cmath.exp(1j * (g + p + l)) * cos_t
                         ]],
                        dtype=complex)


class CXGate(TwoQubitGate):
    """Controlled-X (CNOT) gate"""
    lowername = "cx"
    cu_params = (math.pi, 0.0, math.pi, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'CXGate':
        return cls(targets, params)

    def dagger(self):
        return self

    def matrix(self):
        return np.array(
            [[1, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0]],
            dtype=complex)


class CYGate(TwoQubitGate):
    """Controlled-Y gate"""
    lowername = "cy"
    cu_params = (math.pi, math.pi / 2, math.pi / 2, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'CYGate':
        return cls(targets, params)

    def dagger(self):
        return self

    def matrix(self):
        return np.array(
            [[1, 0, 0, 0], [0, 0, -1j, 0], [0, 0, 1, 0], [0, 1j, 0, 0]],
            dtype=complex)


class CZGate(TwoQubitGate):
    """Controlled-Z gate"""
    lowername = "cz"
    cu_params = (0.0, 0.0, math.pi, 0.0)

    @classmethod
    def create(cls, targets, params) -> 'CZGate':
        return cls(targets, params)

    def dagger(self):
        return self

    def matrix(self):
        return np.array(
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]],
            dtype=complex)


class RXXGate(TwoQubitGate):
    """Rotate-XX gate"""
    lowername = "rxx"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta, ), **kwargs)
        self.theta = theta

    @classmethod
    def create(cls, targets, params) -> 'RXXGate':
        return cls(targets, *params)

    def dagger(self):
        return RXXGate(self.targets, -self.theta, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits, lambda c, t: [
                HGate(c),
                HGate(t),
                RZZGate((c, t), self.theta),
                HGate(c),
                HGate(t)
            ])

    def matrix(self):
        a = math.cos(self.theta * 0.5)
        b = -1j * math.sin(self.theta * 0.5)
        return np.array(
            [[a, 0, 0, b], [0, a, b, 0], [0, b, a, 0], [a, 0, 0, b]],
            dtype=complex)


class RYYGate(TwoQubitGate):
    """Rotate-YY gate"""
    lowername = "ryy"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta, ), **kwargs)
        self.theta = theta

    @classmethod
    def create(cls, targets, params) -> 'RYYGate':
        return cls(targets, *params)

    def dagger(self):
        return RYYGate(self.targets, -self.theta, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits, lambda c, t: [
                RXGate(c, -math.pi * 0.5),
                RXGate(t, -math.pi * 0.5),
                RZZGate((c, t), self.theta),
                RXGate(c, math.pi * 0.5),
                RXGate(t, math.pi * 0.5)
            ])

    def matrix(self):
        a = math.cos(self.theta * 0.5)
        b = 1j * math.sin(self.theta * 0.5)
        return np.array(
            [[a, 0, 0, b], [0, a, -b, 0], [0, -b, a, 0], [a, 0, 0, b]],
            dtype=complex)


class RZZGate(TwoQubitGate):
    """Rotate-ZZ gate"""
    lowername = "rzz"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta, ), **kwargs)
        self.theta = theta

    @classmethod
    def create(cls, targets, params) -> 'RZZGate':
        return cls(targets, *params)

    def dagger(self):
        return RZZGate(self.targets, -self.theta, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits, lambda c, t:
            [CXGate(
                (c, t)), RZGate(t, self.theta),
             CXGate((c, t))])

    def matrix(self):
        a = cmath.exp(0.5j * self.theta)
        return np.array([[a.conjugate(), 0, 0, 0], [0, a, 0, 0], [0, 0, a, 0],
                         [0, 0, 0, a.conjugate()]],
                        dtype=complex)


class SwapGate(TwoQubitGate):
    """Swap gate"""
    lowername = "swap"

    def dagger(self):
        return self

    @classmethod
    def create(cls, targets, params) -> 'SwapGate':
        return cls(targets, params)

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits,
            lambda c, t: [CXGate(
                (c, t)), CXGate(
                    (t, c)), CXGate((c, t))])

    def matrix(self):
        return np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0],
                         [0, 0, 0, 1]])


class ZZGate(TwoQubitGate):
    """ZZ gate

    This gate is a basis two-qubit gate for some kinds of trapped-ion based machines.
    It is equivalent with RZZ(pi/2) except global phase.
    """
    lowername = "zz"

    def __init__(self, targets, **kwargs):
        super().__init__(targets, (), **kwargs)

    @classmethod
    def create(cls, targets, params) -> 'ZZGate':
        return cls(targets)

    def dagger(self):
        return self

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits, lambda c, t: [
                RYGate(t, math.pi * 0.5),
                CXGate((c, t)),
                RZGate(c, math.pi * 0.5),
                UGate(t, -math.pi * 0.5, math.pi * 0.5, 0.0, math.pi * 0.25),
            ])

    def matrix(self):
        return np.diag([1, 1j, 1j, 1])


class Measurement(Operation):
    """Measurement operation"""
    lowername = "measure"

    @classmethod
    def create(cls, targets, params) -> 'Measurement':
        return cls(targets)

    # Ad-hoc copy and paste programming.
    def target_iter(self, n_qubits):
        """The generator which yields the target qubits."""
        return slicing(self.targets, n_qubits)


class Reset(Operation):
    """Reset operation"""
    lowername = "reset"

    @classmethod
    def create(cls, targets, params) -> 'Reset':
        return cls(targets)

    # Ad-hoc copy and paste programming.
    def target_iter(self, n_qubits):
        """The generator which yields the target qubits."""
        return slicing(self.targets, n_qubits)


class DeprecatedOperation:
    """Inform deprecated operation"""
    def __init__(self, name: str, alternative: str) -> None:
        self.name = name
        self.alt = alternative

    def __call__(self, *_args, **_kwargs) -> NoReturn:
        raise ValueError(
            f'{self.name} operation is deprecated. Use insteads {self.alt}.')


def slicing_singlevalue(arg: Union[slice, int], length: int) -> Iterator[int]:
    """Internally used."""
    if isinstance(arg, slice):
        start, stop, step = arg.indices(length)
        i = start
        if step > 0:
            while i < stop:
                yield i
                i += step
        else:
            while i > stop:
                yield i
                i += step
    else:
        try:
            i = arg.__index__()
        except AttributeError:
            raise TypeError("indices must be integers or slices, not " +
                            arg.__class__.__name__) from None
        if i < 0:
            i += length
        yield i


def slicing(args: Targets, length: int) -> Iterator[int]:
    """Internally used."""
    if isinstance(args, tuple):
        for arg in args:
            yield from slicing_singlevalue(arg, length)
    else:
        yield from slicing_singlevalue(args, length)


def qubit_pairs(args: Tuple[Targets, Targets], length: int) -> Iterator[Tuple[int, int]]:
    """Internally used."""
    if not isinstance(args, tuple):
        raise ValueError("Control and target qubits pair(s) are required.")
    if len(args) != 2:
        raise ValueError("Control and target qubits pair(s) are required.")
    controls = list(slicing(args[0], length))
    targets = list(slicing(args[1], length))
    if len(controls) != len(targets):
        raise ValueError(
            "The number of control qubits and target qubits are must be same.")
    for c, z in zip(controls, targets):
        if c == z:
            raise ValueError(
                "Control qubit and target qubit are must be different.")
    return zip(controls, targets)


def get_maximum_index(indices: Targets) -> int:
    """Internally used."""
    def _maximum_idx_single(idx: int):
        if isinstance(idx, slice):
            start = -1
            stop = 0
            if idx.start is not None:
                start = idx.start.__index__()
            if idx.stop is not None:
                stop = idx.stop.__index__()
            return max(start, stop - 1)
        return idx.__index__()

    if isinstance(indices, tuple):
        return max((_maximum_idx_single(i) for i in indices), default=-1)
    return _maximum_idx_single(indices)


def find_n_qubits(gates: Iterable[Operation]) -> int:
    """Find n_qubits from gates"""
    return max((get_maximum_index(g.targets) for g in gates), default=-1) + 1
